/**
 * @file imgutil.c
 * @brief Utilities for image manipulation
 * @author Florian Schmitt (0schmitt@informatik.uni-hamburg.de)
 * @date Jul-2012
**/
#include "camcapture.h"
#include <stdlib.h>

/** @cond Makros for color conversation */
#define YCBCR_Y(r, g, b)  (16  + ((  65.738*r  +  129.057*g  +   25.064*b ) / 256))
#define YCBCR_Cb(r, g, b) (128 + (( -37.945*r  +  -74.494*g  +  112.439*b ) / 256))
#define YCBCR_Cr(r, g, b) (128 + (( 112.439*r  +  -94.154*g  +  -18.285*b ) / 256))
/** @endcond */



/**
 * Extracts YCbCr-values from RGB-colors
 * @param rgb RGB-Input color
 * @param Y  (if nullpointer: not calculated)
 * @param Cb (if nullpointer: not calculated)
 * @param Cr (if nullpointer: not calculated)
**/
void rgb_to_ycbcr(const color_rgb_t *rgb, int *Y, int *Cb, int *Cr)
{
	/* calculate */
	if(Y  != 0) *Y  = (int) YCBCR_Y(  ((double) rgb->r), ((double) rgb->g), ((double) rgb->b) );
	if(Cb != 0) *Cb = (int) YCBCR_Cb( ((double) rgb->r), ((double) rgb->g), ((double) rgb->b) );
	if(Cr != 0) *Cr = (int) YCBCR_Cr( ((double) rgb->r), ((double) rgb->g), ((double) rgb->b) );
}



/**
 * Copy data from image src to image dest. This function just copies them and does not allocate memory.
 * If you want to allocate memory, too, you should use camcapture_clone_image()!
 * @param src  source image
 * @param dest destination image
**/
void camcapture_copy_image(const image_data_t *src, image_data_t *dest)
{
	int n, s1;
	void *sd;
	void *td;
	/* image sizes have to be equal */
	if( (src->width != dest->width) || (src->height != dest->height) )
		return;
	/* size of picture one */
	s1 = dest->width * dest->height / 2;
	/* data pointers */
	sd = src->img;
	td = dest->img;
	/* copy loop */
	for(n=0; n<s1; n++)
		*((int*)td+n) = *((int*)sd+n);
}



/**
 * Mallocates a new image, equal in size and dimmensions to another
 * @param preset  preset image
 * @param dest    destination image
**/
void camcapture_create_compatible_image(const image_data_t *preset, image_data_t *dest)
{
	/* set height and width */
	dest->height = preset->height;
	dest->width  = preset->width;
	/* allocate new image memory now */
	dest->img = malloc((sizeof(int)*dest->width * dest->height)/2);
}



/**
 * Frees the images data.
 * Please notice:  Do never free the image you get by camcapture_get_frame()!
 * @param dest    destination image
**/
void camcapture_free_image_data(image_data_t *dest)
{
	free(dest->img);
}



/**
 * Copy data from image src to image dest. This function frees dests memory and reallocates it automatically.
 * If you just want to copy, you should use camcapture_copy_image()! Dont forget to free memory later!
 * @param src  source image
 * @param dest destination image
**/
void camcapture_clone_image(const image_data_t *src, image_data_t *dest)
{
	int n, s1;
	void *sd;
	void *td;
	/* Allocate a compatible image */
	camcapture_create_compatible_image(src, dest);
	/* size of picture one */
	s1 = dest->width * dest->height / 2;
	/* data pointers */
	sd = src->img;
	td = dest->img;
	/* copy loop */
	for(n=0; n<s1; n++)
		*((int*)td+n) = *((int*)sd+n);
}



/**
 * Get a pixels color
 * @param image Image data struct
 * @param px x coordinate
 * @param px y coordinate
 * @param color to get (rgb)
**/
void camcapture_get_pixel(image_data_t *image, int px, int py, color_rgb_t *color)
{
	int tmp, Y0, Y1, Cb, Cr, value;
	int y0, y1, pb, pr, r, g, b;
	void *data;

	/* calc color and get image pointer */
	data = image->img;

	/* calculate data position into tmp */
	tmp = (py*image->width + px) / 2;
	/* get color */
	value = *((int*)data+tmp);
	Y0 = (int)(value & 0xFF);
	Cb = (int)((value >> 8) & 0xFF);
	Y1 = (int)((value >> 16) & 0xFF);
	Cr = (int)((value >> 24) & 0xFF);

	y0 = (255 / 219.0) * (Y0 - 16);
	y1 = (255 / 219.0) * (Y1 - 16);
	pb = (255 / 224.0) * (Cb - 128);
	pr = (255 / 224.0) * (Cr - 128);

	/* Get pixel */
	if( ((py*image->width + px)) % 2 )
	{
		/* first pixel */
		r = 1.0 * y0 + 0     * pb + 1.402 * pr;
		g = 1.0 * y0 - 0.344 * pb - 0.714 * pr;
		b = 1.0 * y0 + 1.772 * pb + 0     * pr;
	}
	else
	{
		/* second pixel */
		r = 1.0 * y1 + 0     * pb + 1.402 * pr;
		g = 1.0 * y1 - 0.344 * pb - 0.714 * pr;
		b = 1.0 * y1 + 1.772 * pb + 0     * pr;
	}
	/* Save to struct */
	color->r = clamp(r);
	color->g = clamp(g);
	color->b = clamp(b);
}
